# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''数据驱动模块

使用介绍：
1、在TestCase类前面加入@DataDrive装饰器，参数为测试数据列表或者字典::

    @datadrive.DataDrive([1,2,3])
    class Test(tc.TestCase):
        def runTest(self):
            pass
    
    @datadrive.DataDrive(
        {
        'TEST1':1,
        'TEST2':2,
        'TEST3':3,
        }
        )
    class Test(tc.TestCase):
        def runTest(self):
            pass
        
在测试报告中展示用例名字时，若参数为列表的，则在用例后面增加一个索引序号，若参数为字典的，则在用例后面增加上key的名字：

    - 列表显示为::
    
        Test#1
        Test#2
        Test#3
        
    - 字典显示为::

        Test#TEST1
        Test#TEST2
        Test#TEST3

2、可以在runTest里通过self.casedata使用测试数据::

    def runTest(self):
        print(self.casedata)
    
3、运行及调试方法和原来一样::

    MyTest().run()
        

    完整的例子如下::
    
    # -*- coding: utf-8 -*-
    
    import testbase.testcase as tc
    import testbase.datadrive as datadrive
    
    @datadrive.DataDrive([1,2,3])
    class MyTest(tc.TestCase):
        """test
        """
        
        owner = 'foo'
        priority = tc.TestCase.EnumPriority.High
        status = tc.TestCase.EnumStatus.Ready
        timeout = 5       
        
        def runTest(self):
            print 'runTest, casedata:', self.casedata
              
            
    if __name__ == '__main__':
        MyTest().run()
        
'''
from __future__ import absolute_import

import six
import types

from testbase import logger
from testbase.conf import settings
from testbase.testcase import TestCase
from testbase.util import translate_bad_char, has_bad_char, smart_text


class DataDrive(object):
    '''数据驱动类修饰器，标识一个测试用例类使用数据驱动
    '''

    def __init__(self, case_data):
        '''构造函数
        
        :param case_datas: 数据驱动测试数据集
        :type case_datas: list/tuple/dict
        '''
        self._case_data = case_data

    def __call__(self, testcase_class):
        '''修饰器
        
        :param testcase_class: 要修饰的测试用例
        :type testcase_class: TestCase
        '''
        if not issubclass(testcase_class, TestCase):
            raise TypeError('data driver decorator cannot be applied to non-TestCase type')
        testcase_class.__qtaf_datadrive__ = self
        return testcase_class

    def __iter__(self):
        '''遍历全部的数据名
        '''
        if isinstance(self._case_data, types.GeneratorType):
            self._case_data = list(self._case_data)
        if isinstance(self._case_data, list) or isinstance(self._case_data, tuple):
            for it in range(len(self._case_data)):
                yield it
        else:
            for it in self._case_data:
                yield it

    def __getitem__(self, name):
        '''获取对应名称的数据
        '''
        return self._case_data[name]

    def __len__(self):
        return len(self._case_data)


def is_datadrive(obj):
    '''是否为数据驱动用例
    
    :param obj: 测试用例或测试用例类
    :type obj: TestCase/type
    
    :returns boolean
    '''
    return hasattr(obj, '__qtaf_datadrive__')


def get_datadrive(obj):
    '''获取对应用例的数据驱动
    
    :param obj: 测试用例或测试用例类
    :type obj: TestCase/type
    
    :returns DataDrive
    '''
    return obj.__qtaf_datadrive__


def _get_translated_in_datadrive(name, dd):
    if isinstance(name, six.string_types):
        name = smart_text(name)
    else:
        name = str(name)
    translated_name = translate_bad_char(name)
    for item in dd:
        if isinstance(item, six.string_types):
            item_string = smart_text(item)
        else:
            item_string = str(item)
        if translated_name == translate_bad_char(item_string):
            return dd[item]
    else:
        raise ValueError("data drive name '%s' not found" % name)


def load_datadrive_tests(cls, name=None):
    '''加载对应数据驱动测试用例类的数据驱动用例
    '''
    if is_datadrive(cls):
        dd = get_datadrive(cls)
    else:
        if not settings.DATA_DRIVE:
            raise RuntimeError("DATA_DRIVE is not set to True")

        from testbase.loader import TestDataLoader
        dd = TestDataLoader().load()

    if name:
        if name in dd:
            drive_data = {name : dd[name]}
        else:
            drive_value = _get_translated_in_datadrive(name, dd)
            drive_data = {name : drive_value}
    else:
        drive_data = dd

    tests = []
    for item in drive_data:
        testdata = drive_data[item]
        if isinstance(item, six.string_types):
            item = smart_text(item)
        else:
            item = str(item)
        casedata_name = item
        if has_bad_char(item):
            casedata_name = translate_bad_char(item)
            warn_msg = "[WARNING]%r's drive data key should use \"%s\" instead of \"%s\"" % (cls, casedata_name, item)
            logger.warn(warn_msg)

        if isinstance(testdata, dict) and "__attrs__" in testdata:
            attrs = testdata.get("__attrs__")
        else:
            attrs = None
        tests.append(cls(testdata, casedata_name, attrs))
    return tests




# -*- coding: utf-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#
'''
测试用例加载
'''
from __future__ import absolute_import

import imp
import pkgutil
import os
import six
import sys
import traceback
import types

from collections import OrderedDict
from testbase import datadrive
from testbase.testcase import TestCase, SeqTestSuite
from testbase.conf import settings
from testbase.util import smart_text


class TestLoader(object):
    '''测试用例加载器
    '''

    def __init__(self, filter_func=None):
        '''构造函数
        :param filter: 用例过滤函数，函数原型为 filter_func(testcase_obj)，返回None/False表示不过滤此用例
        :type filter: callable
        '''
        self._filter = filter_func
        self._module_errs = {}
        self._filtered_tests = {}

    def get_last_errors(self):
        '''返回最后一次load调用时加载失败的全部模块和对应错误信息
        
        :returns dict: 模块和对应的错误信息
        '''
        return self._module_errs

    def get_filtered_tests(self):
        '''返回最后一次load调用时被过滤掉的测试用例
        '''
        return list(self._filtered_tests.keys())

    def get_filtered_tests_with_reason(self):
        '''返回最后一次load调用时被过滤掉的测试用例和过滤原因
        '''
        return self._filtered_tests

    def load(self, testname):
        '''通过名字加载测试用例
        
        :param name: 用例或用例名称
        :type name: string/list
        
        :returns list - 测试用例对象列表
        '''

        if isinstance(testname, list):
            testnames = testname
        else:
            testnames = [testname]

        self._module_errs = {}
        testcases = []

        for testname in testnames:
            testname = smart_text(testname)
            if settings.DATA_DRIVE:
                self._dataset = TestDataLoader().load()
            if '/' in testname:
                testname, data_key = testname.split('/', 1)
            else:
                data_key = None

            obj = self._load(testname)

            if isinstance(obj, types.ModuleType):
                if hasattr(obj, '__path__'):
                    testcases += self._load_from_package(obj, data_key)
                else:
                    testcases += self._load_from_module(obj, data_key)
            elif isinstance(obj, type):
                testcases += self._load_from_class(obj, data_key)

        # 过滤掉重复的用例
        testcase_dict = {}
        for testcase in testcases:
            testcase_dict[testcase.test_name] = testcase

        return list(testcase_dict.values())

    def _load(self, testname):
        '''加载对应的对象
        
        :param name: 用例或用例集名称
        :type name: string
        :returns - Type/ModuleType
        '''
        parts = testname.split('.')
        module = None
        parts_imp = parts[:]
        while parts_imp:
            try:
                modulename = '.'.join(parts_imp)
                __import__(modulename)  # __import__得到的是最外层模块的object
                module = sys.modules[modulename]
                break
            except ImportError:
                del parts_imp[-1]
            except:
                del parts_imp[-1]
                break

        obj = module

        if parts_imp == parts:  # 为一个包或模块
            return obj

        elif parts_imp == parts[0:-1] and hasattr(module, parts[-1]):  # 为一个类
            try:
                testclass = getattr(module, parts[-1])
                if not self._is_testcase_class(testclass):
                    raise TypeError("%s不是一个有效的测试用例" % testname)
            except:
                self._module_errs[testname] = traceback.format_exc()
                return
            else:
                return testclass

        else:  # 触发异常
            if parts_imp:
                modulename = '.'.join(parts_imp)
                modulename += '.'
            else:
                modulename = ''
            modulename += parts[len(parts_imp)]
            try:
                __import__(modulename)
            except:
                self._module_errs[modulename] = traceback.format_exc()

    def _is_testcase_class(self, obj):
        '''是否为测试用例类
        
        :returns bool - 是否为用例类
        '''
        return (isinstance(obj, type) and
              issubclass(obj, TestCase) and
              hasattr(obj, "runTest") and
              getattr(obj, "priority", None))

    def _walk_package_error(self, modulename):
        '''walk_packages错误回调
        '''
        self._module_errs[modulename] = traceback.format_exc()

    def _load_from_package(self, pkg, data_key=None):
        '''从一个python包加载测试用例
        
        :param pkg: Python包
        :type pkg: ModuleType
        :returns list - 测试用例对象列表
        '''
        tests = []
        for _, modulename, ispkg in pkgutil.walk_packages(pkg.__path__, pkg.__name__ + '.', onerror=self._walk_package_error):
            if ispkg:
                continue
            try:
                __import__(modulename)
                tests += self._load_from_module(sys.modules[modulename], data_key)
            except:
                self._module_errs[modulename] = traceback.format_exc()
        return tests

    def _load_from_module(self, mod, data_key=None):
        '''从一个python模块加载测试用例
        
        :param mod: Python模块
        :type mod: ModuleType
        :returns list - 测试用例对象列表
        '''
        tests = []
        for name in dir(mod):
            obj = getattr(mod, name)
            if self._is_testcase_class(obj):
                tests += self._load_from_class(obj, data_key)
        if hasattr(mod, '__qtaf_seq_tests__'):  # 测试用例需要顺序执行
            seqdef = mod.__qtaf_seq_tests__
            if not isinstance(seqdef, list):
                raise TypeError("__qtaf_seq_tests__必须为list类型")
            if len(seqdef) == 0:
                raise ValueError("__qtaf_seq_tests__必须至少包含一个测试用例")
            for it in seqdef:
                if type(it) != type(TestCase) or not issubclass(it, TestCase):
                    raise TypeError("__qtaf_seq_tests__的元素必须为测试用例类")
            test_dict = {}
            for test in tests:
                test_dict[type(test)] = test
            for it in seqdef:
                if it not in test_dict:
                    raise RuntimeError("__qtaf_seq_tests__中的测试用例'%s'已被过滤" % it.__name__)
            tests = [SeqTestSuite([ test_dict[it] for it in seqdef])]
        return tests

    def _load_from_class(self, cls, data_key=None):
        '''加载用例类
        
        :param cls: Python类 
        :type cls: Type
        :returns list - 测试用例对象列表
        '''
        tests = []
        if datadrive.is_datadrive(cls) or settings.DATA_DRIVE:
            tests = datadrive.load_datadrive_tests(cls, data_key)
        else:
            tests = [cls()]

        if self._filter:
            final_tests = []
            for it in tests:
                filter_reason = self._filter(it)
                if filter_reason:
                    self._filtered_tests[it] = filter_reason
                else:
                    final_tests.append(it)
            tests = final_tests
        return tests


class TestDataLoader(object):
    '''测试数据加载器
    '''
    MODULE_NAME = 'testbase.loader.testdata'

    def _load_dataset_module_from_file(self):
        '''从文件中加载数据模块
        '''
        if self.MODULE_NAME in sys.modules:
            return sys.modules[self.MODULE_NAME]

        if os.path.isfile(__file__):
            qtaf_top_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')
            top_dir = qtaf_top_dir
        else:  # 使用的egg包
            qtaf_top_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')
            top_dir = os.path.join(qtaf_top_dir, '..', '..')
        top_dir = os.path.abspath(top_dir)

        py_path = os.path.join(top_dir, settings.DATA_SOURCE)
        if not os.path.isfile(py_path):
            raise RuntimeError("指定的数据源文件\"%s\"不存在" % py_path)

        module_name = os.path.basename(py_path)
        module_name = module_name[0:module_name.rfind('.')]

        fd, pathname, desc = imp.find_module(module_name, [os.path.dirname(py_path), top_dir])
        module = imp.load_module(self.MODULE_NAME, fd, pathname, desc)
        return module

    def _load_dataset_module_from_name(self):
        '''根据模块名加载模块
        '''
        __import__(settings.DATA_SOURCE)
        return sys.modules[settings.DATA_SOURCE]

    def _load_dataset(self):
        '''加载数据
        '''
        if not settings.DATA_SOURCE:
            raise RuntimeError("DATA_DRIVE=True，但未指定的数据源文件")

        if isinstance(settings.DATA_SOURCE, six.string_types):
            if settings.DATA_SOURCE.endswith('.py'):
                module = self._load_dataset_module_from_file()
            else:
                module = self._load_dataset_module_from_name()
            if not hasattr(module, 'DATASET'):
                raise RuntimeError("数据源文件\"%s\"没有定义模块变量\"DATASET\"" % module)
            return module.DATASET

        else:
            return settings.DATA_SOURCE

    def load(self):
        '''从数据源加载测试数据
        :returns list - 测试数据集
        '''
        dataset = self._load_dataset()
        if isinstance(dataset, list) or isinstance(dataset, tuple):
            dataset_dict = OrderedDict()  # keep ordered
            for idx, it in enumerate(dataset):
                dataset_dict[str(idx)] = it
            dataset = dataset_dict
        elif isinstance(dataset, dict):
            pass
        else:
            raise ValueError("DATASET必须为list/tuple/dict类型")
        if len(dataset) == 0:
            raise ValueError("DATASET不可以为空")
        return dataset
